/*
    PubTrans - Scheduling public transportation on PalmPilot handheld

    Copyright (C) 1999 Andreas Kretzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    e-mail: andi@kretzer-berlin.de
*/

#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>
#include "pubtrans.h"

struct global_db global_db;
UInt times[MAXTIMES];

#define STATION_DBNAME	"Stations"
#define STATION_DBTYPE	'Stat'
#define LINE_DBNAME	"Lines"
#define LINE_DBTYPE	'Line'
#define TLINE_DBNAME	"TimeLines"
#define TLINE_DBTYPE	'TLns'
#define TOUR_DBNAME	"Tour"
#define TOUR_DBTYPE	'Tour'

// #define DEBUG_LIST

static struct dbinit {
   char			*dbname;
   unsigned long	 dbtype;
   DmOpenRef		*db;
} dbinit[] = {
   { STATION_DBNAME, STATION_DBTYPE, &global_db.StationsDB },
   { LINE_DBNAME, LINE_DBTYPE, &global_db.LinesDB },
   { TLINE_DBNAME, TLINE_DBTYPE, &global_db.TlinesDB },
   { TOUR_DBNAME, TOUR_DBTYPE, &global_db.TourDB },
   { NULL, 0, NULL }
};

Boolean OpenDatabase (void)
{
   int error;
   struct dbinit *dbip = dbinit;
   DmOpenRef tmpDB;

   while (dbip && dbip->dbname && *dbip->dbname) {
      error = DmCreateDatabase(0, dbip->dbname, APPID, dbip->dbtype, false);
      if (error && error != dmErrAlreadyExists) return true;
      tmpDB = DmOpenDatabaseByTypeCreator(dbip->dbtype, APPID, dmModeReadWrite);
      *dbip->db = tmpDB;
      if (!tmpDB) return true;
      dbip++;
   }

   return 0;
}

void CloseDatabase (void)
{
   struct dbinit *dbip = dbinit;

   while (dbip && dbip->dbname && *dbip->dbname) {
      if (*dbip->db) DmCloseDatabase (*dbip->db);
      dbip++;
   }
}

Int time_compare (void *p1, void *p2, Int line,
		  SortRecordInfoPtr ignore2, SortRecordInfoPtr ignore3,
		  VoidHand ignore4)
{
   int result;
   UInt tour_line;

   if (line < 0) return -1;
   tour_line = *((UInt *) p2);

   result = 0;
   if ((UInt) line > tour_line) result = 1;
   if ((UInt) line < tour_line) result = -1;
   return result;
}

/*
 * Try to find <station> in <line>. Returns index of the station (0 .. n-1)
 * or -1 if not found.
 */
int station2idx(ULong station, ULong line)
{
   VoidHand rec;
   void *p;
   int i, offset, anz_st;
   UInt *statp;

   if ((rec = DmQueryRecord(global_db.LinesDB, line)) != 0) {
      p = MemHandleLock(rec);
      offset = StrLen(p) + 1;
      if (offset & 1) offset++;		// use only aligned offsets!
      anz_st = (MemHandleSize(rec) - offset) / sizeof(UInt);
      statp = p + offset;
      for (i = 0; i < anz_st; i++) {
	 if (statp[i] == station) break;
      }
      MemHandleUnlock(rec);
      if (i < anz_st) return i;
   }

   return -1;
}

wlist *get_stations(wlist *w)
{
   VoidHand rec;
   void * p;
   int idx = 0;
   char *s, tmp[64];

   if (w) delwlist(w);
   w = genwlist();
   if (!w) return NULL;

   while (w && (rec = DmQueryRecord(global_db.StationsDB, idx)) != 0) {
      p = MemHandleLock(rec);
      MemSet(tmp, 64, 0);
      StrCopy (tmp, p);
      s = uncompress_string(tmp);
#ifdef DEBUG_LIST
      {
	 unsigned long uid;
	 DmRecordInfo(global_db.StationsDB, idx, NULL, &uid, NULL);
	 StrPrintF(tmp, "%lu %s", uid, s);
	 w = addwlist(w, tmp, idx);
      }
#else
      w = addwlist(w, s, idx);
#endif
      MemHandleUnlock(rec);
      idx++;
   }

   return w;
}

struct line *get_lines_by_station(ULong stationID)
{
   VoidHand ln_rec, st_rec;
   struct line *ret, *l, **pp;
   UInt *linep;
   void *p, *st;
   int i, anz_ln, offset;

   ret = l = 0;
   pp = &ret;

   if ((ln_rec = DmQueryRecord(global_db.StationsDB, stationID)) == 0)
   	return NULL;

   p = MemHandleLock(ln_rec);
   offset = StrLen(p) + 1;
   if (offset & 1) offset++;		// use only aligned offsets!

   anz_ln = (MemHandleSize(ln_rec) - offset) / sizeof(UInt);
   linep = p + offset;
   for (i = 0; i < anz_ln; i++) {
      l = MemPtrNew(sizeof(struct line));
      if (l == NULL) {
	 MemHandleUnlock(ln_rec);
	 FrmCustomAlert(alertID_info, "get_lines_by_station()",
				      "No memory for struct line.", "");
	 return ret;
      }
      MemSet(l, sizeof(struct line), 0);
      *pp = l;
      pp = &l->next;

      l->idx = linep[i];
      if ((st_rec = DmQueryRecord(global_db.LinesDB, l->idx)) != 0) {
	 st = MemHandleLock(st_rec);
	 l->name = save_string(uncompress_string(st), 0);
	 MemHandleUnlock(st_rec);
      }
   }
   MemHandleUnlock(ln_rec);

   return ret;
}

void free_timetables (struct line *l)
{
   struct depart *d;

   while (l) {
      while (l->dp) {
         d = l->dp;
         l->dp = d->next;
         MemPtrFree (d);
      }
      l = l->next;
   }
}

void free_lines(struct line *l)
{
   struct line *tmp;

   free_timetables(l);
   while (l) {
      if (l->name) MemPtrFree(l->name);
      tmp = l->next;
      MemPtrFree (l);
      l = tmp;
   }
}

Int find_first_tour (UInt line)
{
   Int idx;
   idx = DmFindSortPosition(global_db.TourDB, NULL, NULL,
   						time_compare, line - 1);
   return idx;
}

static void get_times(struct line *l, int statidx, UInt day)
{
   VoidHand rec, Tl_rec;
   struct depart **p;
   UInt idx, Line_idx, Tline_idx, Time, Days, depart, *vals;
   int t_idx, anz_times;
   void *p1, *p2;

   if (!l) return;
   idx = find_first_tour(l->idx);
   p = &l->dp;
   while ((rec = DmQueryRecord(global_db.TourDB, idx)) != 0) {
      p1 = MemHandleLock(rec);
      vals = (UInt *) p1;
      Line_idx = vals[0];
      if (Line_idx != l->idx) {
	 MemHandleUnlock(rec);
	 break;
      }

      Days = vals[1];
      Tline_idx = vals[2];
      Time = vals[3];
      if (Days & (1 << day)) {
	 if ((Tl_rec = DmQueryRecord(global_db.TlinesDB, Tline_idx)) != 0) {
	    p2 = MemHandleLock(Tl_rec);
	    vals = (UInt *) p2;
	    anz_times = MemHandleSize(Tl_rec) / sizeof(UInt) - 1;
	    t_idx = statidx - vals[1];
	    if (t_idx >= 0 && t_idx < anz_times) {
	       if (t_idx == 0) depart = Time;
	       else depart = Time + vals[t_idx + 1];
	       if ((*p = MemPtrNew(sizeof(struct depart))) != NULL) {
		  MemSet(*p, sizeof(struct depart), 0);
		  (*p)->time = depart;
		  p = &(*p)->next;
	       }
	    }
	    MemHandleUnlock(Tl_rec);
	 }
      }

      MemHandleUnlock(rec);
      idx++;
   }
}

static int build_table_steps(struct line *lp)
{
   int count;
   struct line *l1, *l2;
   struct depart *dp1, *dp2;

   MemSet (times, 0, sizeof(times));
   if (!lp) return 0;

   count = 0;

   l1 = lp;
   l2 = lp->next;

   dp1 = l1->dp;
   dp2 = NULL;
   if (l2) dp2 = l2->dp;

   while (count < MAXTIMES && (dp1 || dp2)) {
      if (dp1 && dp2) {
	 if (dp1->time <= dp2->time) {
	    times[count++] = dp1->time;
	    dp1 = dp1->next;
	    if (dp1 && dp1->time > dp2->time) dp2 = dp2->next;
	 } else {
	    times[count++] = dp2->time;
	    dp2 = dp2->next;
	    if (dp2 && dp2->time > dp1->time) dp1 = dp1->next;
	 }
      } else {
         if (dp1) {
            times[count++] = dp1->time;
            dp1 = dp1->next;
	 } else {
            times[count++] = dp2->time;
            dp2 = dp2->next;
	 }
      }
   }
   return count;
}

struct line *get_timetables(ULong station, short *maxcol)
{
   struct line *lp = NULL;
   struct line *l;

   l = lp = get_lines_by_station(station);

   *maxcol = 0;
   while (l) {
      l = l->next;
      (*maxcol)++;
   }
   return lp;
}

short calc_ruler (struct line *l, int column, ULong station, UInt day)
{
   int col, entries, statidx;
   struct line *l1;

   for (col = 0; col < column && l; col++, l = l->next) ;

   if (l->dp == NULL) {
      statidx = station2idx(station, l->idx);
      if (statidx >= 0) get_times (l, statidx, day);
   }
   l1 = l->next;
   if (l1 && l1->dp == NULL) {
      statidx = station2idx(station, l1->idx);
      if (statidx >= 0) get_times (l1, statidx, day);
   }

   entries = build_table_steps(l);
   return (short) entries;
}

void table_setvals(struct line *l, UInt time, int column, TablePtr table)
{
   struct depart *dp;
   int col, row, maxrow;

   maxrow = TblGetNumberOfRows(table);

   for (col = 0; col < column && l; col++, l = l->next) ;

   for (; col < column + 2 && l; col++, l = l->next) {
      TblSetItemPtr(table, 0, col - column, l->name);
      dp = l->dp;
      row = 1;
      while (dp && row < maxrow) {
	 if (dp->time >= time) {
	    TblSetItemInt (table, row, col - column, dp->time);
	    row++;
	 }
	 dp = dp->next;
      }
   }
}
